//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svg_text_position.h"
#include "svg_styleset.h"
namespace svg
{
//-------------------------------------------------------------------------------	
	//Applies some style attributes to the current text position
	//Arguments:
	//	req:	a request - to obtain the attribute data;
	//	style:	an effective (accumulated from the parents) style,
	//			which should contain the font definition
	//It also modifies the style: scales the stroke-width attribute 
	//to match the internal transform of the glyph vertex source!
	void text_position::apply(request &req, styleset &style)
	{
		if(style.has(attribute::kerning))
		{
			m_kerning_value = style.of(attribute::kerning).val.real;
			m_auto_kerning = (m_kerning_value < 0);//negative kerning means "auto"
		}

		if(style.has(attribute::writing_mode))
			m_wmode = style.of(attribute::writing_mode).val.writing_mode;

		if(style.has(attribute::text_anchor))
			m_anchor = style.of(attribute::text_anchor).val.real;

		if(style.has(attribute::font))
			m_glyph = (font *) req.get_element(style.of(attribute::font).val.id,"ext:font");

		if (m_glyph)
		{
			if(style.has(attribute::font_size))
				m_scale = style.of(attribute::font_size).val.real / m_glyph->size();
			
			if(style.has(attribute::stroke_width) && m_scale != 0.0)
			{
				double width = style.of(attribute::stroke_width).val2double()/m_scale;
				style.set(svg::attribute(attribute::stroke_width,width));
			}

			m_valid = true;

		}
		else
		{
			m_valid = false;
		}
	}//apply
//-------------------------------------------------------------------------------	

	//Sets the initial current text position
	//Arguments:
	//	x,y:		coordinates of the starting position (for the text 
	//				on a curve the "y" coordinate is ignored)
	//	text:		the text, which fragment will be drawn;
	//	pos:		position of the first character that belonggs to this
	//				text fragment (text chunk)
	//	count:		number of characters to write (length of the text chunk)
	//	dx,dy:		offset from current text position, that has to be applied
	//				to each character within this text chunk
	//	rotation:	additional rotation, applied individually to each character
	//				within this text chunk
	void text_position::set(double x, double y, string &text, unsigned pos, 
						unsigned count, double dx, double dy, double rotation)
	{
		m_text = &text;
		m_base = pos;
		m_pos = 0;
		m_count = count;
		//count oversized?:
		if (text.size() < (m_base + m_count)) m_count = text.size() - m_base;
		m_local.reset();
		m_local.rotate(rotation);
		m_local.translate(dx, dy);
		if (m_glyph)
		{
			m_glyph->select(code(m_pos)); //select the first character
		}
	}//set

//-------------------------------------------------------------------------------	
	//offset X to the new character position:
	double text_position::advance_x()
	{
		if (!m_glyph) return 0; //can be called, even when m_glyph is null... 
		switch(m_wmode)
		{
		case rl:
		case lr:
				return (m_glyph->advance_x() ) * m_scale;
		case tb:
			return 0;
		default:
			return 0;
		}//switch
	}//advance_x
//-------------------------------------------------------------------------------	
	//offset Y to the new character position:
	double text_position::advance_y()
	{
		rectangle box = m_glyph->bounds(); 
		switch(m_wmode)
		{
		case rl:
		case lr:
			return 0;
		case tb:
				return m_glyph->size() * 1.11 * m_scale; //absolute experimental:
														 //just differs from Inkscape of
														 //about 1%
		default:
			return 0;
		}//switch
	}//advance_y
//-------------------------------------------------------------------------------	
	//It returns an offset from the reference point: 
	//	- for "left-right": baseline (y=0), on left edge of the glyph box;
	//	- for "top-bottom": middle of the top edge of the glyph box;
	//into natural center of the glyph (identified as the point 
	// 0,0 within the glyph bounds).
	transform text_position::internal_transform()
	{
		transform t;
		rectangle box = m_glyph->bounds(); 
		switch(m_wmode)
		{
		case rl:
		case lr:
			t.translate (
							kerning_x(),
							kerning_y()
						);
			break;
		case tb:
			t.translate (
							 - 0.5*(box.x1 + box.x2) + kerning_x(), 
							 0.75 * m_glyph->size() + kerning_y()//x1,y1 is the upper - left corner 
						);
			break;
		}//switch
		t.scale(m_scale); 
		return t;
	}
//-------------------------------------------------------------------------------	
	//Total text position offset for selected text chunk
	rectangle text_position::offset()
	{
		rectangle result(0,0,0,0);
		
		unsigned i;
		
		unicode actual = m_glyph->selected(); //we will have to restore it later
		m_glyph->select(code(0),false);

		for (i = 1; i < m_count; i++)
		{
			result.x2 += advance_x();
			result.y2 += advance_y();

			m_glyph->select(code(i),(i>0));

			//Compared with Word and Inkscape: strange enough, auto kerning is 
			//not used to position the whole text. It updates the internal
			//glyph position  w i t h i n  its box! - see internal_transformation()
			if (!m_auto_kerning)
			{
				result.x2 += kerning_x(); 
				result.y2 += kerning_y();
			}
		}//for

		result.x2 += advance_x();
		result.y2 += advance_y();

		m_glyph->select(actual,(m_pos > 0)); //go back to the actual character

		return result;
	}//size

} // namespace svg
