//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svgtex.h"
#include "tex_tileset.h"
#include "tex_instance.h"

namespace tex
{
	//returns the tile, which contains requested pixel. It creates a tile
	//"on the fly", when a new one is required.
	//Arguments:
	//	level:		determines resolution of the picture, which is needed.
	//				(level = 8, => resolution = 256x256,
	//				 level = 9, => resulution = 512x512)
	//	x,y:		pixel coordinates (expressed for full image resolution)
	//	inst:		instance object, required to initialize a new tile.
	tile &tileset::get_tile(int32u level, int32 x, int32 y, instance &inst)
	{
		int32u col = x/tile_size;
		int32u row = y/tile_size;
		int32u index;
	
		if (level > m_max_level) grow_to(level);

		if(level <= tile_level)
		{
			index = base_index(level);
		}
		else
		{
			int32u stride = 1 << (level - tile_level);
			index = base_index(level) + (stride * row) + col;
		}
		
		tile &t = m_tiles[index];

		if(!t.is_enabled())
		{
			MSG4("> creating tile #%d res=%d [%d %d]\n",index, (1<<level), col, row);
			t.initialize(inst, level, col, row);

			//update statistical fields:
			m_tile_memory += t.memory();
			if (t.is_enabled()) m_count_enabled++;
			if (t.is_constant_color()) m_count_mono++;
		}
		
		return t.reference(this->m_timestamp++);
	}

	//Returns index of a first tile, older than given age (i.e. his age() < age)
	//Arguments:
	//	age:	age to compare (in the "timestamp" units)
	//When there is no such tiles: returns -1.
	int32 tileset::oldest(int age)
	{
		if (m_tiles.size() < 1 || (m_count_enabled - m_count_mono) < 1)
		{
			return -1; //no tiles, at all!
		}
		
		int32u i;
		int32  index = -1;
		
		for (i = 0; i < m_tiles.size(); i++)
		{
			tile &t = m_tiles[i];
			if (t.is_enabled() && ! t.is_constant_color() && t.age() < age)
			{
				age = t.age();
				index = i;
			}
		}

		return index;
	}

	//Verifies the memory, used by this tileset with the requested maximum.
	//When tileset uses more than the given amount - the oldest tiles are
	//reduced.
	//Additionally, it "conserves" the m_timestamp value, reducing it on all 
	//tiles (to not allow to grow it to infinity)
	//Arguments:
	// megabytes_limit: limit, in megabytes
	void tileset::limit(int32u megabytes_limit)
	{
		//0. Do not check too often!
		if ((m_timestamp - m_checked_timestamp) < tileset::min_check_interval) 
		{
			return;
		}
		else
		{
			m_checked_timestamp = m_timestamp;
		}

		//1. check the memory limit
		int32u maxmem = megabytes_limit << 20; //memory limit, in bytes
		int32  min_age = m_timestamp;
		
		if (m_tiles.byte_size() > (maxmem / 2))  //When over 50% of the max. memory
		{								    //is used just by the "framework"and
											// constant tiles:
			reset();					    //- clear the whole thing!
			MSG0("- tileset has been reset to the empty state\n");
			return;
		}

		while (m_tile_memory > maxmem)
		{
			int32 old = oldest(min_age);
			if(old >= 0)
			{
				tile & t = m_tiles[old];
				m_tile_memory -= t.memory();
				m_count_enabled --;

				MSG1("- deleting tile #%d\n",old);
				m_tiles[old].reset(); 
			}
			else
			{
				break; //no tiles to remove!
			}
		}//end while
		MSG3("- memory now=%d MB, tiles=%d, solid=%d\n", used_memory(),
													m_count_enabled,m_count_mono);

		//2. reduce the timestamp value:
		if (m_timestamp > tileset::max_age)
		{	//keep the timestamp in healthy state:
			maintain_timestamp();
			MSG1(" - timestamp decreased to %d\n",m_timestamp);
		}

	}

	//timestamp maintenence (to not exceed the int range!)
	void tileset::maintain_timestamp()
	{
		if (m_timestamp > max_age)
		{
			int32 old = oldest(m_timestamp);
			if (old < 0) return; //nothing to do...
			int32u decrease = m_tiles[old].age() - 1;
			int32u i;
			for (i = 0; i < m_tiles.size(); i++)
			{
				tile & t = m_tiles[i];
				if(t.is_enabled() && ! t.is_constant_color())
				{
					t.age(t.age()-decrease);
				}//end if
			}//for
			m_timestamp -= decrease;
		}//end if
	}
}//end namespace tex
